def import_profiler_registers():
    import realhf.search_engine.enumerate
    import realhf.search_engine.estimate
    import realhf.search_engine.layers
    import realhf.search_engine.param_realloc
    import realhf.search_engine.search
    import realhf.search_engine.utils
